#!/usr/bin/env python

from rnalib.struct.pairs_util import accuracy
from rnalib.struct.cl_util import compare_classifications

def ens_accuracy(ens, ref_pairs):
    """Return dict {struct_id: acc_dict}
    
    ens -- dict {struct_id: {'P':...,'V':...,...}
    ref_pairs -- Pairs object, reference structure
    """
    result = {}
    for rec_id, info in ens.items():
        result[rec_id] = accuracy(ref_pairs, info['P'])
    return result

def ens_cl_accuracy(ens, ref_cl):
    """Return dict {struct_id: classification_similarity}

    ens -- dict {struct_id: {'P':...,'V':...,...}
    ref_cl -- Classification object, reference classification

    """
    result = {}
    for rec_id, info in ens.items():
        result[rec_id] = compare_classifications(ref_cl, info['C'])
    return result

def acc_single_pairs(pairs, ref_pairs):
    """Return accuracy of a single Pairs object

    pairs -- Pairs object (or list of tuples)
    """
    return accuracy(ref_pairs, pairs)

def cs_single_cl(cl, ref_cl):
    """Return CS (classification similarity) of a single Classification object

    cl, ref_cl -- Classification object
    """
    return compare_classifications(ref_cl, cl)

def write_acc(id_acc, output_filename, known_labels=['SEN','PPV','CC','CS']):
    """Produce tab-delimited file

    id_acc -- dict of {struct_id: {'CC':0.873}}

    format:
    struct_id\tSEN:x.xxx\tCC:x.xxxx\n
    """

    out = open(output_filename, 'w')
    for struct_id in id_acc:
        parts = [struct_id]
        for label in known_labels:
            parts.append("%s:%f"%(label,id_acc[struct_id][label]))
        out.write("%s\n"%('\t'.join(parts)))
    out.close()
    return None

def read_acc(lines):
    """Return dict of {struct_id: {'CC':0.835, 'SEN':0.334}}
    """
    result = {}
    for line in lines:
        parts = line.strip().split('\t')
        struct_id, rest = parts[0], parts[1:]
        acc = {}
        for item in rest:
            k,v = item.split(':')
            acc[k] = float(v)
        result[struct_id] = acc
    return result
