#!/usr/bin/env python

from sys import argv
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.core.moltype import RNA
from cogent.core.sequence import ModelSequence
from numpy import array

from spunclib.seqs_aln import get_aln_info, write_aln_fasta, parse_aln_fasta
from spunclib.structures import parse_struct

BASE_COLORS = {\
    'U':"#FF6600",\
    'C':"#FF9900",\
    'A':"#6699FF",\
    'G':"#66CCFF",\
    }

PAIR_COLORS = {\
    ('G','C'):"#3366FF",\
    ('C','G'):"#3366FF",\
    ('A','U'):"#6699FF",\
    ('U','A'):"#6699FF",\
    ('G','U'):"#66CCFF",\
    ('U','G'):"#66CCFF",\
    }

UNP_COLOR = "#CCCCCC"
MISMATCH_COLOR = "#FFCC00"

def legend_base():
    """Return legend info for base colors

    Returned value is list of tuples (label, color).
    """
    result = []
    for b in "UCAG":
        result.append((b, BASE_COLORS[b]))
    result.append(('Other', UNP_COLOR))
    return result

def legend_pairs():
    """Return legend info for base pair colors

    Returned value is list of tuples (label, color).
    """
    result = []
    result.append(('Unpaired',UNP_COLOR))
    for p in zip("GAG","CUU"):
        x,y = p[0],p[1]
        label = "%s/%s, %s/%s pair"%(x,y,y,x)
        result.append((label,PAIR_COLORS[p]))
    result.append(('Mismatch', MISMATCH_COLOR))
    return result

def get_colors_base(aln):
    """Return list of lists containing colors

    Returned value is basically a matrix of num_seqs x aln_len
    containing a color at each position.
    Cells are colored by residue (BASE_COLORS).
    """
    result = []
    for s in aln.Seqs:
        result.append([])
        for c in range(aln.SeqLen):
            result[-1].append(UNP_COLOR)
    for seq_idx, seq in enumerate(aln.Seqs):
        for char_idx, seq_char in enumerate(seq):
            try:
                c = BASE_COLORS[seq_char]
                result[seq_idx][char_idx] = c
            except KeyError:
                continue
                #c = MISMATCH_COLOR
    return result

def get_colors_pairs(aln, ref):
    """Return list of lists containing colors

    Returned value is basically a matrix of num_seqs x aln_len
    containing a color at each position.
    Cells are colored by pairs (PAIRS_COLORS).
    """
    result = []
    for s in aln.Seqs:
        result.append([])
        for c in range(aln.SeqLen):
            result[-1].append(UNP_COLOR)
    for seq_idx, seq in enumerate(aln.Seqs):
        for x,y in ref:
            try:
                c = PAIR_COLORS[(seq[x],seq[y])]
            except KeyError:
                c = MISMATCH_COLOR
            result[seq_idx][x] = c
            result[seq_idx][y] = c
    return result

def colored_html(aln, ref=None, output_filename=None):
    """Return HTML string describing colored alignment and legend

    aln -- Alignment or DenseAlignment object
    ref -- Pairs object
    output_filename -- string ending in .html, path to output file
    """
    
    col_ids = range(aln.SeqLen)
    if ref is not None:
        colors = get_colors_pairs(aln, ref)
    else:
        colors = get_colors_base(aln)
    
    out_elem = []
    out_elem.append("<html><body>")
    
    # LEGEND
    out_elem.append("<h2>Legend</h2>")
    out_elem.append("<table>")
    if ref is None:
        legend_items = legend_base()
    else:
        legend_items = legend_pairs()
    for label, item_color in legend_items:
        out_elem.append("<tr><td bgcolor=%s><b>%s</b></td></tr>"\
            %(item_color,label))
    out_elem.append("</table>")

    # ALIGNMENT DATA
    out_elem.append("<h2>Alignment data</h2>")
    out_elem.append("""<table border=0 cellpadding=2 cellspacing=0>""")
    # COLUMN IDS
    out_elem.append("<tr>")
    for item in col_ids:
        out_elem.append("<td align=center>%s</td>"%item)
    out_elem.append("<td>Alignment column</td>")
    out_elem.append("</tr")

    if ref is not None:
        ref_partners = ref.toPartners(aln.SeqLen)
        # PAIRING PARTNERS
        out_elem.append("<tr>")
        for item in ref_partners:
            if item is not None:
                out_elem.append("<td align=center>%s</td>"%item)
            else:
                out_elem.append("<td></td>")
        out_elem.append("<td>Pairing column</td>")        
        out_elem.append("</tr>")

    # SEQ INFO
    for seq_idx, seq in enumerate(aln.Seqs):
        out_elem.append("<tr>")
        for char_idx, seq_char in enumerate(seq):
            out_elem.append("<td align=center bgcolor=%s>%s</td>"\
                %(colors[seq_idx][char_idx],seq_char))
        out_elem.append("<td nowrap>%s</td>"%(seq.Name))
        out_elem.append("</tr>")
    
    out_elem.append("</table>")
    out_elem.append("</body></html>")
    
    out_str = '\n'.join(out_elem) 
    
    if output_filename is not None:
        out = open(output_filename,'w')
        out.write(out_str)
        out.close()
    return out_str

if __name__ == "__main__":

    aln = parse_aln_fasta(open(argv[1]), dense=True)
    try:
        ref_pairs = parse_struct(argv[2])
    except IndexError:
        ref_pairs = None
    colored_html(aln, ref_pairs, output_filename='test_aln.html')
