#!/usr/bin/env python
# ensemble.py

"""
Regulated version. Each method returns a max of 200 structures.

Author: Sandra Smit (Sandra.Smit@colorado.edu)
Revision History:
File created on 06 Nov 2007.

"""
from __future__ import division
from numpy import arange
from random import sample
from os import makedirs
from optparse import OptionParser
#from rnalib.format.input import get_aln_pairs
from rnalib.struct.classifier import SLBO
from rnalib.struct.structgen import mfe_structs_aln, subopt_structs_aln,\
    degrade_helices, extend_helices, get_unique_pairs, keep_percentage,\
    add_bp, add_pr, mfe_constrained, subopt_constrained, mfe_structs,\
    subopt_structs
from rnalib.struct.pairs_util import pairs_to_mask, vienna_mask,\
    pairs_to_mask_regions
from cogent.core.moltype import RNA
from cogent.parse.fasta import MinimalFastaParser
from cogent.core.alignment import Alignment, DenseAlignment
from cogent.core.sequence import ModelSequence, RnaSequence
from cogent.struct.pairs_util import *
from cogent.struct.knots import *
from rnalib.struct.constraint import *

CONSTRAINTS = [constraint_A, constraint_B, constraint_C, constraint_D,\
    constraint_E, constraint_F]
PKREM = [inc_order, inc_range, inc_length, elim_conf, elim_gain]

def get_aln_pairs(lines, dense=False, alpha=RNA.Alphabets.DegenGapped):
    """Return tuple of Alignment and Pairs object
    
    lines -- file stream or list of lines
    dense -- bookean, when False Alignment is returned, when True
        DenseAlignmentis returned. Default is False.

    Handles format:
    >consensus_vienna
    ...((..))...
    >consensus_mask
    ...11..11...
    >seq1
    ACCGGUACCUGC
    >seq2
    A--GCUUGC-GA
    ...
    """
    
    cons_vienna = None
    cons_mask = None
    cons_pairs = None
    ref_seq = None
    seqs = []
    aln = None

    for label, seq in MinimalFastaParser(lines):
        if label == 'consensus_vienna':
            cons_vienna = seq
            continue
        elif label == 'consensus_mask':
            cons_mask = seq
            continue
        else:
            if label.startswith('REFSEQ*'):
                ref_seq = RNA.Sequence(seq, Name=label)
            if dense:
                s = ModelSequence(seq, Name=label, Alphabet=alpha)
            else:
                s = RnaSequence(seq, Name=label)
            seqs.append(s)

    if dense:
        aln = DenseAlignment(seqs, MolType=RNA, Alphabet=alpha)
    else:
        aln = Alignment(seqs, MolType=RNA)
    cons_pairs = vienna_mask(cons_vienna, cons_mask)

    return aln, cons_pairs, ref_seq


def method_A(aln, pairs, ref_seq=None):
    """mfe structures
    """
    result = []
    vienna_structs = mfe_structs_aln(aln, degap=True)
    for v in vienna_structs:
        result.append(v.toPairs())
    #if len(result) > 200:
    #    return sample(result, 200)
    #else:
    return result

def method_B(aln, pairs, ref_seq=None):
    """subopt structures
    """
    sub_structs = subopt_structs_aln(aln, energy_range=2, degap=True)
    nolp_structs = subopt_structs_aln(aln, energy_range=2, noLP=True,\
        degap=True)
    result = []
    for v in sub_structs+nolp_structs:
        result.append(v.toPairs())
    return result

def method_C(aln, pairs, ref_seq=None):
    """Sample from suboptimal structures"""
    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 1000 #number of wanted structures
    er = 5 # energy range
    num_per_seq = int(num_of_structs/num_seqs)
    structs = subopt_structs_aln(aln, degap=True,\
        sample_ensemble=num_per_seq, energy_range=er)
    result = []
    for v in structs:
        result.append(v.toPairs())
    #if len(result) > 200:
    #    return sample(result, 200)
    #else:
    return result

def method_D(aln, pairs, ref_seq=None):
    """Degrade helices
    """
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    l = len(degap_pairs)
    two_thirds = int((2/3)*l)
    print two_thirds
    step = int(two_thirds/10)
    if step < 1:
        step = 1
    print step
    print range(1,two_thirds,step)

    result = []
    for x in range(10): # 20 structures
        for num in range(1,two_thirds,step):
            new_struct = degrade_helices(degap_pairs, num)
            gapped_seq, gapped_pairs = ungapped_to_gapped(ref_seq, new_struct)
            result.append(gapped_pairs)
    return result

def method_E(aln, pairs, ref_seq=None):
    """Keep certain percentage"""
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    result = []
    
    for frac in arange(0.4,1.0,0.05):
        new_structs1 = keep_percentage(degap_pairs, fraction=frac,\
            regions=False,number=10)
        new_structs2 = keep_percentage(degap_pairs, fraction=frac,\
            regions=True, number=10)
        for p in new_structs1+new_structs2:
            gapped_seq, gapped_pairs = ungapped_to_gapped(ref_seq, p)
            result.append(gapped_pairs)
    return result

def method_F(aln, pairs, ref_seq=None):
    """Extend helices
    """
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    l = len(degap_pairs)
    two_thirds = int((2/3)*l)
    print two_thirds
    step = int(two_thirds/10)
    if step < 1:
        step = 1
    print step
    print range(1,two_thirds,step)

    result = []
    for x in range(10): # 20 structures
        for num in range(1,20,2):
            new_struct = extend_helices(degap_pairs, seq_len=aln.SeqLen,\
                number=num)
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq, new_struct)
            result.append(gapped_struct)
    return result

def method_F2(aln, pairs, ref_seq=None):
    """add pr, bp"""
    assert aln.SeqLen == len(ref_seq)
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    result = []
    degap_prs = PairedRegionsFromPairs(degap_pairs)
    for percentage in arange(0.1, 0.45, 0.05):
        bp_number = int(percentage*len(degap_pairs)) or 1
        pr_number = int(percentage*len(degap_prs)) or 1
        for x in range(10):
            new_struct1 = add_bp(degap_pairs, len(degap_seq), number=bp_number)
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq, new_struct1)
            assert gapped_seq == ref_seq
            result.append(gapped_struct)
            new_struct2 = add_pr(degap_pairs, len(degap_seq), number=pr_number)
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq, new_struct2)
            result.append(gapped_struct)
    return result

def method_G(aln, pairs, ref_seq=None):
    """constrained folding full range, MFE structs"""
    assert aln.SeqLen == len(ref_seq)
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    
    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 50 #number of wanted structures
    
    result = []
    for x in range(50):
        new_structs = []
        chosen_pk_rem = choice(PKREM)
        chosen_constraint = choice(CONSTRAINTS)
        chosen_fraction = choice(arange(0.05,1,0.05))
        pkfree_pairs = chosen_pk_rem(degap_pairs)
        con = chosen_constraint(pkfree_pairs, len(degap_seq),\
            frac=chosen_fraction)
        # calculate mfe structure under constraint
        new_structs.extend(mfe_constrained([(degap_seq, con)]))

        for vienna in new_structs:
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq,\
                vienna.toPairs())
            result.append(gapped_struct)

    return result

def method_H(aln, pairs, ref_seq=None):
    """constrained folding full range"""
    #assert aln.SeqLen == len(ref_seq)
   
    if pairs is None:
        raise ValueError("Need input pairs for this method")

    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 1000 #number of wanted structures
    sample_size = 5
    er = 5 # energy range
    num_per_seq = int(num_of_structs/(num_seqs*sample_size))

    if num_per_seq < 1:
        num_per_seq = 1
    ns = aln.NamedSeqs

    result = []
    for seq_obj in aln.Seqs:
        seq_str = str(seq_obj)
        degap_seq, degap_pairs = gapped_to_ungapped(seq_str, pairs)
        for x in range(num_per_seq):
            new_structs = []
            # select pseudoknot removal, constraint, fraction
            chosen_pk_rem = choice(PKREM)
            chosen_constraint = choice(CONSTRAINTS)
            chosen_fraction = choice(arange(0.05,1,0.05))
            # create nested structure and constraint
            pkfree_pairs = chosen_pk_rem(degap_pairs)
            con = chosen_constraint(pkfree_pairs, len(degap_seq),\
                frac=chosen_fraction)
            # sample x number of structures
            new_structs.extend(subopt_constrained([(degap_seq, con)],\
                sample_ensemble=sample_size))
            #print 'sample done'

            for vienna in new_structs:
                #print vienna
                gapped_seq, gapped_struct = ungapped_to_gapped(seq_str,\
                    vienna.toPairs())
                result.append(gapped_struct)
    #if len(result) > 200:
    #    return sample(result, 200)
    #else:
    return result


def method_I(aln, pairs, ref_seq=None):
    """constrained folding 50-100% extend helices"""
    assert aln.SeqLen == len(ref_seq)
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    result = []
    for x in range(50):
        new_structs = []
        chosen_pk_rem = choice(PKREM)
        chosen_constraint = choice(CONSTRAINTS)
        chosen_fraction = choice(arange(0.5,1.01,0.02))
        pkfree_pairs = chosen_pk_rem(degap_pairs)
        con = chosen_constraint(pkfree_pairs, len(degap_seq),\
            frac=chosen_fraction)
        new_structs.extend(mfe_constrained([(degap_seq, con)]))
        new_structs.extend(subopt_constrained([(degap_seq, con)],\
            sample_ensemble=5))
        #print 'sample done'

        for vienna in new_structs:
            new_pairs = vienna.toPairs()
            new_prs = PairedRegionsFromPairs(new_pairs)
            to_extend = int(choice(arange(0.1, 0.51, 0.05))*len(new_prs))
            ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
            #print vienna
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq,\
                ext_pairs)
            result.append(gapped_struct)
    if len(result) > 200:
        return sample(result, 200)
    else:
        return result



def method_J(aln, pairs, ref_seq=None):
    """constrained folding 70-100% extend helices"""
    assert aln.SeqLen == len(ref_seq)
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    result = []
    for x in range(20):
        new_structs = []
        chosen_pk_rem = choice(PKREM)
        chosen_constraint = choice(CONSTRAINTS)
        chosen_fraction = choice(arange(0.7,1.01,0.02))
        pkfree_pairs = chosen_pk_rem(degap_pairs)
        con = chosen_constraint(pkfree_pairs, len(degap_seq),\
            frac=chosen_fraction)
        new_structs.extend(subopt_constrained([(degap_seq, con)],\
            sample_ensemble=5))
        #print 'sample done'

        for vienna in new_structs:
            new_pairs = vienna.toPairs()
            new_prs = PairedRegionsFromPairs(new_pairs)
            to_extend = int(choice(arange(0.1, 0.51, 0.05))*len(new_prs))
            ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
            #print vienna
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq,\
                ext_pairs)
            result.append(gapped_struct)
    if len(result) > 200:
        return sample(result, 200)
    else:
        return result


def method_X(aln, pairs, ref_seq=None):
    """constrained folding 70-100% extend helices"""
    num_seqs = len(aln.Seqs) # number of seqs
    num_of_structs = 1000 # requested number of structures
    num_per_seq = int(num_of_structs/num_seqs)
    result = []
    for seq_to_fold in aln.Seqs:
        degap_seq, degap_pairs = gapped_to_ungapped(seq_to_fold, pairs)
        chosen_fraction = choice(arange(0.1,0.52,0.05))
        con = constraint_unp(len(degap_seq),chosen_fraction)
        new_structs = []
        new_structs.extend(subopt_constrained([(degap_seq, con)],\
            sample_ensemble=num_per_seq))
        for vienna in new_structs:
            new_pairs = vienna.toPairs()
            gapped_seq, gapped_struct = ungapped_to_gapped(seq_to_fold,\
                new_pairs)
            result.append(gapped_struct)
    return result

    


def parse_command_line_parameters():
    """ Parses command line arguments """
    result = OptionParser()

    # A binary 'verbose' flag
    result.add_option('-v','--verbose',action='store_true',dest='verbose')
    result.add_option('-m','--method',action='store',\
                      type='string',dest='method')
    result.add_option('-c','--cl_function',action='store',\
                      type='string',dest='cl_function')
    result.add_option('-p','--prefix',action='store',\
                      type='string',dest='prefix')
    
    result.add_option('-R','--real',action='store',\
                      type='string',dest='real')
    result.add_option('-o','--outdir',action='store',\
                      type='string',dest='outdir')

    
    # Set default values here if they should be other than None
    result.set_defaults(verbose=False)
    result.set_defaults(method=None)
    result.set_defaults(cl_function=SLBO)
    result.set_defaults(prefix='structures')
    return result.parse_args()

def method_M(aln, pairs, ref_seq=None):
    """MFE structs, extend helices
    """
    
    num_seqs = len(aln.Seqs) #number of seqs
    #num_of_structs = 1000 #number of wanted structures
    #num_per_seq = int(num_of_structs/(num_seqs*sample_size))

    result = []
    for seq_obj in aln.Seqs:
        new_structs = []
        seq_str = str(seq_obj)
        degap_seq, degap_pairs = gapped_to_ungapped(seq_str, pairs)

        new_structs.extend(mfe_structs([degap_seq]))

        for vienna in new_structs:
            new_pairs = vienna.toPairs()
            new_prs = PairedRegionsFromPairs(new_pairs)
            to_extend = int(choice(arange(0.1, 0.95, 0.05))*len(new_prs))
            ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
            print len(ext_pairs)/len(new_pairs)
            gapped_seq, gapped_struct = ungapped_to_gapped(seq_str,\
                ext_pairs)
            result.append(gapped_struct)

    return result

def method_N(aln, pairs, ref_seq=None):
    """MFE structs, extend helices
    """
    
    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 1000 #number of wanted structures
    er = 5 # energy range
    sample_size = 5
    num_per_seq = int(num_of_structs/(num_seqs*sample_size))
    if num_per_seq < 1:
        num_per_seq = 1

    result = []
    for seq_obj in aln.Seqs:
        seq_str = str(seq_obj)
        degap_seq, degap_pairs = gapped_to_ungapped(seq_str, pairs)
        for x in range(num_per_seq):
            new_structs = []
            # sample x number of structures
            new_structs.extend(subopt_structs([degap_seq],\
                sample_ensemble=sample_size, energy_range=er))

            for vienna in new_structs:
                new_pairs = vienna.toPairs()
                new_prs = PairedRegionsFromPairs(new_pairs)
                to_extend = int(choice(arange(0.1, 0.95, 0.05))*len(new_prs))
                ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
                gapped_seq, gapped_struct = ungapped_to_gapped(seq_str,\
                    ext_pairs)
                result.append(gapped_struct)

    return result


def method_O(aln, pairs, ref_seq=None):
    """constrained folding full range, MFE structs, extend helices"""
    assert aln.SeqLen == len(ref_seq)
    degap_seq, degap_pairs = gapped_to_ungapped(ref_seq, pairs)
    
    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 50 #number of wanted structures
    
    result = []
    for x in range(num_of_structs):
        new_structs = []
        chosen_pk_rem = choice(PKREM)
        chosen_constraint = choice(CONSTRAINTS)
        chosen_fraction = choice(arange(0.05,1,0.05))
        pkfree_pairs = chosen_pk_rem(degap_pairs)
        con = chosen_constraint(pkfree_pairs, len(degap_seq),\
            frac=chosen_fraction)
        # calculate mfe structure under constraint
        new_structs.extend(mfe_constrained([(degap_seq, con)]))

        for vienna in new_structs:
            new_pairs = vienna.toPairs()
            new_prs = PairedRegionsFromPairs(new_pairs)
            to_extend = int(choice(arange(0.1, 0.95, 0.05))*len(new_prs))
            ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
            gapped_seq, gapped_struct = ungapped_to_gapped(ref_seq,\
                ext_pairs)
            result.append(gapped_struct)

    return result

def method_P(aln, pairs, ref_seq=None):
    """constrained folding full range, helix extension"""
    #assert aln.SeqLen == len(ref_seq)
    
    num_seqs = len(aln.Seqs) #number of seqs
    num_of_structs = 1000 #number of wanted structures
    sample_size = 5
    er = 5 # energy range
    num_per_seq = int(num_of_structs/(num_seqs*sample_size))

    if num_per_seq < 1:
        num_per_seq = 1
    ns = aln.NamedSeqs

    result = []
    for seq_obj in aln.Seqs:
        seq_str = str(seq_obj)
        degap_seq, degap_pairs = gapped_to_ungapped(seq_str, pairs)
        for x in range(num_per_seq):
            new_structs = []
            # select pseudoknot removal, constraint, fraction
            chosen_pk_rem = choice(PKREM)
            chosen_constraint = choice(CONSTRAINTS)
            chosen_fraction = choice(arange(0.05,1,0.05))
            # create nested structure and constraint
            pkfree_pairs = chosen_pk_rem(degap_pairs)
            con = chosen_constraint(pkfree_pairs, len(degap_seq),\
                frac=chosen_fraction)
            # sample x number of structures
            new_structs.extend(subopt_constrained([(degap_seq, con)],\
                sample_ensemble=sample_size, energy_range=er))

            for vienna in new_structs:
                new_pairs = vienna.toPairs()
                new_prs = PairedRegionsFromPairs(new_pairs)
                to_extend = int(choice(arange(0.1, 0.95, 0.05))*len(new_prs))
                ext_pairs = extend_helices(new_pairs, len(degap_seq), to_extend)
                gapped_seq, gapped_struct = ungapped_to_gapped(seq_str,\
                    ext_pairs)
                result.append(gapped_struct)

    return result


if __name__ == "__main__":

    METHODS = {'A': method_A, 'B': method_B, 'C': method_C, 'D': method_D,\
                'E': method_E, 'F': method_F, 'G': method_G, 'H': method_H,\
                'I': method_I, 'J': method_J, 'M':method_M, 'N':method_N,\
                'O': method_O, 'P': method_P}

    opts,arg = parse_command_line_parameters()

    if not opts.real:
        print "Provide alignment file"
        exit()
    
    aln_filename = opts.real
    print 'ALIGNMENT FILE: %s'%(aln_filename)

    
    if opts.outdir:
        outdir = opts.outdir
    else:
        outdir = "./"
    if not outdir.endswith("/"):
        outdir += "/"
    try:
        makedirs(outdir)
    except OSError: # if already exists, ignore
        pass


    aln, cons_pairs, ref_seq = get_aln_pairs(open(aln_filename), dense=True)

    if opts.method is None:
        methods_to_run = METHODS.keys()
        methods_to_run.sort()
    else:
        methods_to_run = opts.method
    print "METHODS TO RUN: %s"%(methods_to_run)

    for method_label in methods_to_run:
        # STRUCTURE GENERATION
        gen_method = METHODS[method_label]
        all_pairs = gen_method(aln, cons_pairs, ref_seq=ref_seq)
        unique_pairs = get_unique_pairs(all_pairs)

        # OUTPUT
        out_filename = outdir + '.'.join([opts.prefix, method_label,'ens'])
        out = open(out_filename, 'w')
        for idx, pairs in enumerate(unique_pairs):
            key = method_label + '_struct_'+str(idx)
            v, m = pairs_to_mask_regions(pairs, aln.SeqLen, recycle=True)
            assert len(v) == aln.SeqLen
            c = opts.cl_function(pairs, aln.SeqLen)
            frac_s = c.count('S')/len(c)
            if frac_s < 0.2:
                continue
            out.write('>'+key+'\n')
            out.write(str(v)+'\n')
            out.write(str(m)+'\n')
            out.write(str(c)+'\n')
        out.close()
