#!/usr/bin/env python
# ncprop.py

"""
Author: Sandra Smit (S.Smit@few.vu.nl)

Revision History:
File created on 19 Feb 2008.

"""
from __future__ import division
from numpy import array, sqrt

def simple_prop(struct_elem, comp_axis, prop):
    """
    struct_elem: string as used in the dictionary, e.g. S,L,B,O,E,F
    comp_axis: two-letter axis specification: UC, UG, or UA
    prop: int, 0=standard deviation, 1=mean, 2=min, 3=max
    """
    def apply_to(ncstat):
        try:
            return ncstat[struct_elem][comp_axis][prop]
        except KeyError:
            return None
    return apply_to

STD_S_UC = simple_prop('S','UC', 0)
STD_S_UG = simple_prop('S','UG', 0)
MEAN_S_UA = simple_prop('S','UA', 1)

def avg_prop(struct_elems, comp_axes, prop):

    def apply_to(ncstat):
        all_numbers = []
        for ss in struct_elems:
            for ax in comp_axes:
                try:
                    all_numbers.append(ncstat[ss][ax][prop])
                except KeyError:
                    continue
        filtered = filter(None, all_numbers)
        if filtered:
            rec_value = array(filtered).mean() #average(filtered)
        else:
            rec_value = None
        return rec_value
    return apply_to

STD_LBO_UX = avg_prop("LBO", ['UC','UG','UA'], 0)
MEAN_LBO_UA = avg_prop('LBO', ['UA'], 1)
MEAN_BO_UA = avg_prop('BO', ['UA'], 1)
STD_L_UX = avg_prop("L", ['UC','UG','UA'], 0)
STD_B_UX = avg_prop("B", ['UC','UG','UA'], 0)
STD_O_UX = avg_prop("O", ['UC','UG','UA'], 0)
STD_LB_UX = avg_prop("LB", ['UC','UG','UA'], 0)

def prop_single(ncstat, prop_f):
    """Return value of property

    ncstats -- dict containing nucleotide stats for single structure
    """
    return prop_f(ncstat)

def prop_multi(ncstats, prop_f):
    """Return mu and sigma from distribution of property in ensemble
    """
    prop_values = []
    for rec_id, data in ncstats.items():
        prop_values.append(prop_f(data))
    prop_values = array(filter(None, prop_values))
    if not prop_values.any():
        return None, None
    prop_mean = prop_values.mean()
    prop_std = prop_values.std()
    return prop_mean, prop_std

def prop_z_score(value, mu, sigma):
    """Return Z-score = (value-mu)/sigma

    value = value of property
    mu = mean of distribution of property values
    sigma = standard deviation of distribution of property values
    """
    return (value-mu)/sigma


class CompProp(object):
    """Compositional Property
    """

    def __init__(self, PropId, PropFunc, RefZ=None, RefMean=None, Weight=1):
        self.PropId = PropId
        self.PropFunc = PropFunc
        self.RefZ = RefZ
        self.RefMean = RefMean
        self.Weight = Weight

def score_function(comp_props):
    """
    prop_functions -- list of property functions. A prop function can 
        be applied to a ncstat dict and returns a value or None
    """
    def apply_to(ncstats):
        
        sd_per_member = {} # squared differences
        for cp in comp_props:
            ens_mean, ens_std = prop_multi(ncstats, cp.PropFunc)
            for member, ms in ncstats.items(): # member_id, member_stats
                if member not in sd_per_member:
                    sd_per_member[member] = {}
                member_val = prop_single(ms, cp.PropFunc)
                if member_val and ens_mean and ens_std:
                    member_z = prop_z_score(member_val, ens_mean, ens_std)
                    squared_diff = cp.Weight * ((member_z - cp.RefZ)**2)
                    sd_per_member[member][cp.PropId] = squared_diff
        rm_per_member = {} # root mean 
        for member, prop_scores in sd_per_member.items():
            vals = array([i for i in prop_scores.values()])
            rm_per_member[member] = sqrt(vals.sum()/len(vals))
        
        #count = 0
        #for k in sd_per_member:
        #    print 'SD', sd_per_member[k]
        #    print 'RMSD', rm_per_member[k]
        #    count += 1
        #    if count == 10:
        #        break
        #print per_member
        #print mean_root_member

        return rm_per_member


        
    return apply_to


if __name__ == "__main__":
     pass
     #S1 = {'S':{'UC','UG','UA':[
