#!/usr/bin/env python
# predict.py

"""
Author: Sandra Smit (S.Smit@few.vu.nl)

Revision History:
File created on 09 Jun 2008.

"""
# python built-in
from __future__ import division
from sys import exit
from optparse import OptionParser
from os import getcwd, walk
# cogent
# rnalib
from rnalib.struct.classifier import SLBO
from rnalib.struct.pairs_util import vienna_mask, pairs_to_mask, accuracy,\
    pairs_to_mask_regions, bp_fractions, bp_consensus

# local modules
from spunclib.seqs_aln import get_aln_info, write_aln_fasta, parse_aln_fasta
from spunclib.ensemble import write_ensemble, parse_ensemble, generate_ensemble
from spunclib.stats import get_stats, write_stats, parse_stats, check_ens_stats
from spunclib.score import create_score_function, get_scores,\
    SF_WE, SF_WT,SF_UE, SF_UT, read_scores, write_scores, write_top,\
    get_top_scoring
from spunclib.structures import parse_struct, write_struct, get_bp_consensus
from spunclib.accuracy import ens_accuracy, ens_cl_accuracy, acc_single_pairs,\
    write_acc
from spunclib.general import absolute_dir, absolute_file


CL_METHOD = SLBO
EVAL_CHOICES = {'UE': SF_UE, 'UT': SF_UT, 'WE': SF_WE, 'WT': SF_WT}


def parse_command_line_parameters():
    """ Parses command line arguments """
    usage = "usage: %prog [options] aln_file"
    result = OptionParser(usage=usage)

    defaults = {'--verbose': False,
                '--store': False,
                '--bp_cutoff': 0.5,
                '--top_cutoff': 0.1,
                '--eval_choice': 'WE'}

    help_messages =\
        {'--bp_cutoff':\
        "Base pair cutoff for consensus calculation [default=%.3f]"\
        %defaults['--bp_cutoff'],
        '--store': "if True: store alignment, ensemble, stats, and output"}

    # A binary 'verbose' flag
    result.add_option('-v','--verbose',action='store_true',dest='verbose')
    result.add_option('-s','--store',action='store_true',dest='store',\
        help=help_messages['--store'])

    # An example string option
    result.add_option('-E','--ens_filename',action='store',\
                      type='string',dest='ens_filename')
    result.add_option('-S','--stats_filename',action='store',\
                      type='string',dest='stats_filename')
    result.add_option('-R','--ref_struct',action='store',\
                      type='string',dest='ref_struct')
    result.add_option('-b','--bp_cutoff',action='store',\
                      type='float',dest='bp_cutoff',\
                      help=help_messages['--bp_cutoff'])
    result.add_option('-t','--top_cutoff',action='store',\
                      type='float',dest='top_cutoff')
    result.add_option('-e','--eval_choice',action='store',\
                      type='string',dest='eval_choice')
    

    # Set default values here if they should be other than None
    result.set_defaults(verbose=defaults['--verbose'])
    result.set_defaults(bp_cutoff=defaults['--bp_cutoff'])
    result.set_defaults(top_cutoff=defaults['--top_cutoff'])
    result.set_defaults(eval_choice=defaults['--eval_choice'])

    return result.parse_args()


#def get_top_scoring(ens, scores, top_cutoff=0.1, min_struct=10):
#    """Return list of IDs of top-scoring structures
#    """
#    score_id = [(j,i) for i,j in scores.items()]
#    score_id.sort() # from low to high, low=best
#    num_to_select = int(top_cutoff*len(score_id))
#    if num_to_select < min_struct:
#        num_to_select = min_struct
#    top_structs = [rec_id for (score, rec_id) in score_id[:num_to_select]]
#    return top_structs

#def get_bp_consensus(ens, scores, top_cutoff=0.1, bp_cutoff=0.5, min_struct=10):
#    score_id = [(j,i) for i,j in scores.items()]
#    score_id.sort() # from low to high, low=best
#    num_to_select = int(top_cutoff*len(score_id))
#    if num_to_select < min_struct:
#        num_to_select = min_struct
#    top_structs = [ens[rec_id]['P'] for (score, rec_id) in\
#                    score_id[:num_to_select]]
#    bp_fracs = bp_fractions(top_structs)
#    bp_cons = bp_consensus(bp_fracs, cutoff=bp_cutoff)
#    return bp_cons

#def check_ens_stats(ens, stats):
#    """Check same length and keys
#    """
#    if not len(ens) == len(stats):
#        raise ValueError("Ensemble and Stats have different lengths")
#    ens_keys = ens.keys()
#    ens_keys.sort()
#    stats_keys = stats.keys()
#    stats_keys.sort()
#    if ens_keys != stats_keys:
#        raise ValueError("Statistics should have same keys as Ensemble")
    

def predict(aln_filename, ens_filename=None, stats_filename=None,\
    ref_filename=None, output_dir=None, output_basis="test_spunc_test",\
    store=True,\
    bp_cutoff=0.5, top_cutoff=0.1, eval_choice="WE", visualize=False):

    if output_dir is not None:
        output_dir = absolute_dir(output_dir)
    else:
        output_dir = getcwd()

    #aln_filename = args[0]
    if aln_filename is not None:
        aln_filename = absolute_file(aln_filename)
    else:
        raise ValueError("Aln file not specified")

    # 1) ALIGNMENT
    #aln = get_aln_info(aln_filename) #(Dense)Alignment object
    aln = parse_aln_fasta(open(aln_filename), dense=True)
    #(Dense)Alignment object
    
    
    # 2) ENSEMBLE AND STATS
    if ens_filename is None: # if not specified
        ens = generate_ensemble(aln, cl_method=CL_METHOD)
        #format id_pvmc {rec_id:{'P':, 'V', }}
    else:
        ens_filename = absolute_file(ens_filename)
        ens = parse_ensemble(ens_filename, strict=True)
    
    if stats_filename is None: # if not specified
        stats = get_stats(aln, ens)
    else:
        stats_filename = absolute_file(stats_filename)
        stats = parse_stats(stats_filename)
    
    check_ens_stats(ens, stats)

    # 3) SCORE ENSEMBLE MEMBERS
    scoring_function = EVAL_CHOICES[eval_choice]
    scores = get_scores(stats, scoring_function)
    top_scoring = get_top_scoring(ens, scores, top_cutoff=top_cutoff)
    #print top_scoring
    #print len(top_scoring)

    # 4) GET BP CONSENSUS
    topcon = get_bp_consensus(ens, scores, top_cutoff=top_cutoff,\
        bp_cutoff=bp_cutoff)
    enscon = get_bp_consensus(ens, scores, top_cutoff=1.0,\
        bp_cutoff=bp_cutoff)

    #print len(aln), len(aln.Seqs)
    #print len(ens)
    #print len(stats)
    #print len(scores)
    #print topcon
    #print type(topcon)
    #print enscon

    # 5) LOAD REFERENCE STRUCTURE
    ref_pairs = None
    if ref_filename is not None:
        ref_struct = absolute_file(ref_filename)
        ref_pairs = parse_struct(ref_struct)
        ref_cl = SLBO(ref_pairs, aln.SeqLen)

    # ACCURACY
    if ref_pairs:
        id_acc = ens_accuracy(ens, ref_pairs)
        id_cl_acc = ens_cl_accuracy(ens, ref_cl)
        for struct_id in id_cl_acc:
            id_acc[struct_id]['CS'] = id_cl_acc[struct_id]
        enscon_acc = acc_single_pairs(enscon, ref_pairs)
        topcon_acc = acc_single_pairs(topcon, ref_pairs)

    if store:
        write_aln_fasta(aln, output_dir+output_basis+'.aln')
        write_ensemble(ens, output_dir+output_basis+'.ens')
        write_stats(stats, output_dir+output_basis+'.stats')
        write_scores(scores, output_dir+output_basis+'.scores')
        write_top(top_scoring, output_dir+output_basis+'.top')
        write_struct(enscon, aln.SeqLen, output_dir+output_basis+'.enscon')
        write_struct(topcon, aln.SeqLen, output_dir+output_basis+'.topcon')
        if ref_pairs:
            write_struct(ref_pairs, aln.SeqLen, output_dir+output_basis+'.ref')
            write_acc(id_acc, output_dir+output_basis+'.acc')


    
        


    # VISUALIZE
    if visualize:
        from spunclib.visual import plot_rank_vs_score,\
            plot_ens_acc, plot_ens_top, plot_score_acc
        this_fig1 = plot_rank_vs_score(scores,\
            output_dir+output_basis+'.fig.rs.pdf')
        if ref_pairs:
            this_fig2 = plot_ens_acc(id_acc, id_cl_acc,\
                output_dir+output_basis+'.fig.ens.pdf')
            this_fig3 = plot_ens_top(id_acc, top_scoring, enscon_acc,\
                topcon_acc,\
                output_filename=output_dir+output_basis+'.fig.et.pdf')
            this_fig4 = plot_score_acc(id_acc, scores, acc_metric='CC',\
                output_filename = output_dir+output_basis+'.fig.as.pdf')
        #show()

    # 5) RETURN RESULTS
    return scores, topcon


if __name__ == "__main__":

    #opts,args = parse_command_line_parameters()
    # expecting options and one argument (alignment file)
    if len(args) != 1:
        exit()


    s, c = predict(opts=opts, args=args)
    
    
